import os.path
import os
import subprocess
import zipfile
import xml.etree.ElementTree as ET
import saspy
import logging
import tempfile
from ..globals import all_obj_types
from sys import platform


logging.getLogger(__name__)

if platform == "linux" or platform == "linux2":
    from ..globals import java_linux as java
else:
    from ..globals import java_windows as java


def create_saspyconn(sasobjsp_host, sasobjsp_port, appserver, java=java):
    saspy_conn = tempfile.NamedTemporaryFile(mode='w+', delete=False)  # needed to manually deleted after, to be compitable with win
    saspy_conn.write(f"""SAS_config_names=['sasconnection']
SAS_config_options = {{'lock_down': False,
                    'verbose'  : True,
                    'prompt'   : True
                    }}
SAS_output_options = {{'output' : 'html5'}} 
sasconnection = {{'java'      : '{java}',
            'iomhost'   : ['{sasobjsp_host}'],
            'iomport'   : {sasobjsp_port},
            'appserver' : '{appserver}',
            'encoding': 'cyrillic',
            'options' : ["-fullstimer"]
            }}""")
    saspy_conn.seek(0)
    saspy_conn.close()
    return saspy_conn


class AbstractSAS(saspy.SASsession):
    def __init__(self, sasobjsp_user,
                       sasobjsp_pass,
                       sasobjsp_host,
                       sasobjsp_port,
                       appserver,
                       java=java,
                       **kwargs):
        logging.debug('Initializing Abstract SAS class')
        self.sasobjsp_host = sasobjsp_host
        self.sasobjsp_port = sasobjsp_port
        self.sasobjsp_user = sasobjsp_user
        self.sasobjsp_pass = sasobjsp_pass
        self.appserver = appserver
        self.java = java
        logging.debug('Generating saspy conn file')
        self.saspy_conn = create_saspyconn(sasobjsp_host=self.sasobjsp_host,
                                           sasobjsp_port=self.sasobjsp_port,
                                           appserver=self.appserver,
                                           java=self.java)
        logging.debug(f'Generated saspy conn file {self.saspy_conn.name}')
        super(AbstractSAS, self).__init__(cfgfile=self.saspy_conn.name,
                                           cfgname='sasconnection',
                                           omruser=self.sasobjsp_user,
                                           omrpw=self.sasobjsp_pass)
        print(f'Pid in sas server side({self.sasobjsp_host}) is  {self.SASpid}')

    def read_dataset(self, tablename: str, libname: str = 'WORK'):
        """
        Read data From SAS. Data is returned is in the form of Pandas DataFrame.
        :param libname: library, is which table is stored
        :param tablename: name of the table, you want to read
        :return: pd.Dataframe
        """
        df = self.sasdata2dataframe(table=tablename, libref=libname)
        return df

    def write_dataset(self, df, tablename: str, libname: str = 'WORK'):
        """
        Writes dataframe To SAS.
        :param df: dataframe to write
        :param tablename: table, in which you want to write data.
        :param libname: library, in which you want to write data.
        :return:
        """
        result = self.dataframe2sasdata(df=df, table=tablename, libref=libname)
        return result

    def show_libs(self):
        libs = self.assigned_librefs()
        return libs

    def run_script(self, script):
        if script:
            pass
        return self.submit(code=script)

    def endsas(self):
        os.unlink(self.saspy_conn.name)  # need to do it manually for win systems
        super(AbstractSAS, self).endsas()


def get_sas_connect(sasobjsp_host, sasobjsp_port, sasobjsp_user, sasobjsp_pass, java=java):
    """
    LEGACY!!!!!
    :param sasobjsp_host:
    :param sasobjsp_port:
    :param sasobjsp_user:
    :param sasobjsp_pass:
    :param java:
    :return:
    """
    saspy_conn = tempfile.NamedTemporaryFile(mode='w+', delete=True)
    saspy_conn.write(f"""SAS_config_names=['sasconnection']
SAS_config_options = {{'lock_down': False,
                    'verbose'  : True,
                    'prompt'   : True
                    }}
SAS_output_options = {{'output' : 'html5'}} 
sasconnection = {{'java'      : '{java}',
            'iomhost'   : ['{sasobjsp_host}'],
            'iomport'   : {sasobjsp_port},
            'encoding': 'cyrillic',
            'options' : ["-fullstimer"]
            }}""")
    saspy_conn.seek(0)
    try:
        sas = saspy.SASsession(cfgfile=saspy_conn.name, omruser=sasobjsp_user,
                               omrpw=sasobjsp_pass)
    except Exception as e:
        print(f'error!!! {str(e)}')
        return
    return sas


def get_objects_from_spk(spk_path: str, log_path: str = None, object_types: list = None) -> dict:
    """
    Extracts information about metadata objects from spk
    :param spk_path: path (str) to spk
    :param log_path: path(str) to log file
    :param object_types: list of objects types (str), needed to be extracted.
    If not provided - defaults to ['DeployedFlow', 'DeployedJob',
    'ExternalFile', 'Folder', 'Library', 'Role', 'Server', 'StoredProcess',
    'Table', 'User']
    :return: dict with keys: 'Success' - indicating the success of operation and
    'Objects', containing dict with specified metadata object keys, containing
    list of path's --- FIX THAT DESCRIPTION!!!
    """
    logging.info(f"Object extraction from {spk_path} started.")
    log_lines = []
    success_status = False
    if not object_types:
        object_types = all_obj_types
    extracted_objects = {i: [] for i in object_types}
    logging.debug(f"Objects to extract: {extracted_objects}")
    logging.debug(f"Opening file {spk_path}")
    try:
        spk = zipfile.ZipFile(spk_path)
        if 'DeployMap.xml' in spk.namelist():
            deploymap = ET.parse(spk.open('DeployMap.xml'))
            root = deploymap.getroot()
            for obj_type in extracted_objects.keys():
                for item in root.findall(f'Objects/{obj_type}'):
                    obj_path = item.get('Path').replace(f'({obj_type})', '')
                    extracted_objects[obj_type].append(obj_path)
                    logging.debug(f"Founded object {obj_type} in {obj_path}")
                    log_lines += f"Extracted object {obj_type}: {obj_path}\n"
            success_status = True
            logging.info(f"List of objects extracted successfully.")
    except Exception as e:
        logging.error(f"Errors during extraction: {str(e)}")
        log_lines += str(e)
    finally:
        logging.debug(f"Writing log to {log_path}")
        if log_path:
            with open(log_path, 'a') as logfile:
                logfile.writelines(log_lines)
        return {'Success': success_status, 'Objects': extracted_objects}


def check_sas_code(sas_script_path: str = None,
                   sas_script:str = None,
                   log_path: str = None) -> dict:
    """
    Sas code checking. Implements various sas scripth checks
    (drop database includes, etc.)
    :param sas_script_path: path to sas log file (str)
    :param log_path: path to validation log file (str)
    :return: dict with key 'Success', indicating the results of validation.
    """
    logging.info(f"Checking of script {sas_script_path} stated.")
    validation_log_lines = []
    validation_status = False
    try:
        if sas_script_path:
            with open(sas_script_path, 'r') as f:
                sas_script = f.read()
                ### checks....
                pass
        elif sas_script:
            pass
        else:
            raise ValueError
        validation_log_lines += 'Validated\n'
        validation_status = True
        logging.info(f"Script validated successfully.")
    except Exception as e:
        validation_log_lines += str(e)
        logging.critical(f"Script did not passed validation.")
    finally:
        logging.debug(f"Writing log to {log_path}")
        with open(log_path, 'a') as logfile:
            logfile.writelines(validation_log_lines)
        return {'Success': validation_status}


def import_spk(sas_conn_file_path, spk_path, log_path, import_package_jar_path,
               target='/',
               safe_mode=True):
    """
    Importing or checking import of spk to target platform
    :param sas_conn_file_path: path (str) to sas .swa connection profile of
    target platform
    :param spk_path: path(str) to spk
    :param log_path: path (str) to logfile
    :param target: metadata root path (str) to import data. Defaults to '/',
    rarely changes to something else.
    :param import_package_jar_path: path (str) to
    ImportPackage or ImportPackage.exe If not provided - platform and global vars
    are used to determine which path to use.
    :param safe_mode: (bool) - Flag, indicating if spk is being imported,
    or just checking (-noexecute flag)
    :return: dict with key 'Success', results of import or check, key 'Warnings',
    indicating if there were warnings during the process
    """
    logging.info(f"Import of {spk_path} started.")
    logging.debug(f"'Check mode' is set to {safe_mode}")
    execution_log_lines = []
    execution_status = False
    warnings = False
    command = ['-profile', sas_conn_file_path, '-package', spk_path, '-target',
               target, '-log', log_path, '-preservePaths', ]
    if platform == "linux" or platform == "linux2":
        command.append('-disableX11')
    command.insert(0, import_package_jar_path)
    if safe_mode:
        logging.debug(f"Launching in -noexecute mode.")
        command.append('-noexecute')
    try:
        logging.debug(f"'Command to be run: {command}")
        import_process = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if import_process.returncode == 0:
            execution_status = True
            logging.info(f"Import process finished successfully")
        if import_process.returncode == 4:
            execution_status = True
            logging.warning(f"Import process finished with Warnings")
            warnings = True
    except Exception as e:
        execution_log_lines += str(e)
        logging.error(f"Import process finished ERRORS! {str(e)}")
    finally:
        logging.info(f"Log will be written to {log_path}")
        with open(log_path, 'a') as logfile:
            logfile.writelines(execution_log_lines)
        return {'Success': execution_status, 'Warnings': warnings}


def check_import(sas_conn_file_path,
                 spk_path,
                 import_package_jar_path,
                 log_path,
                 target='/'):
    """
    Spk import check, calls import_spk with key "safe_mode=True".
    Writes logs to provided log file.
    :param sas_conn_file_path: path (str) to sas .swa connection profile of
    target platform
    :param spk_path: path(str) to spk
    :param log_path: path (str) to logfile
    :param target: metadata root path (str) to import data. Defaults to '/',
    rarely changes to something else.
    :param import_package_jar_path: path (str) to
    ImportPackage \ ImportPackage.exe If not provided - platform and global vars
    are used to determine which path to use.
    :return: dict with key 'Success', results of import \ check, key 'Warnings',
    indicating if there were warnings during the process
    """
    logging.info(f"Import check of {spk_path} started.")
    check_result = import_spk(sas_conn_file_path=sas_conn_file_path,
                              spk_path=spk_path,
                              log_path=log_path,
                              target=target,
                              import_package_jar_path=import_package_jar_path,
                              safe_mode=True)
    return check_result


def launch_sas_code(sasobjsp_host: str,
                    sasobjsp_port: str,
                    sasobjsp_user: str,
                    sasobjsp_pass: str,
                    sas_script: str = None,
                    sas_script_path: str = None,
                    sas_appserver: str = 'SASApp - Workspace Server',
                    log_path: str = None):
    if sas_script and sas_script_path:
        logging.error(f"You cannot specify 'sas script' an 'sas_script_path' simultaneously")
        raise SyntaxError
    execution_status = False
    execution_log_lines = []
    try:
        if sas_script_path:
            logging.info(f"Launching of script {sas_script_path} started.")
            with open(sas_script_path, 'r') as f:
                sas_script = f.read()
                execution_log_lines += 'The following script obtained from file:\n'
                execution_log_lines += sas_script
        logging.info(f"Starting sas session...")
        sas = AbstractSAS(sasobjsp_user=sasobjsp_user, sasobjsp_pass=sasobjsp_pass,
                          sasobjsp_host=sasobjsp_host, sasobjsp_port=sasobjsp_port, appserver=sas_appserver)
        code = sas.submit(sas_script)
        sas.endsas()
        execution_log_lines += code['LOG']
        execution_status = True
    except Exception as e:
        logging.error(f"Error while executing script. {str(e)}")
        execution_log_lines += str(e)
    finally:
        logging.info(f"Log will be written to {log_path}")
        if log_path:
            with open(log_path, 'a') as log:
                log.writelines(execution_log_lines)
        return {'Success': execution_status}


def redeploy_jobs(objects_list,
                  serverusername,
                  serverpassword,
                  execution_log_path,
                  sas_conn_file_path,
                  deploy_dir,
                  deployjob_jar_path,
                  deploytype='REDEPLOY',
                  metarepository='Foundation',
                  appservername='SASApp',
                  display='localhost:99'):
    """
    Launches redeploy or deploy process of given objects on given platform.
    Writes results to log.
    :param objects_list: (list) of path's (str) of metata objects, being
    deployed or redeployed
    :param serverusername: (str) username for app server
    :param serverpassword: (str) password for app server
    :param execution_log_path: path (str) to log file
    :param sas_conn_file_path: path (str) to sas .swa connection file
    :param deploytype: DEPLOY or REDEPLOY (str), indicates the type of action.
    :param deploy_dir: (str) - path where to deploy .sas code
    :param metarepository: (str) Name of repository, where deploy or redeploy is
    happening. Defaults to 'Foundation'.
    :param appservername: (str) name of App server, to perform action. Defaults
    to 'SASApp'.
    :param deployjob_jar_path: path (str) to DeployJobs \ DeployJobs.exe If not
    provided - platform and global vars
    are used to determine which path to use.
    :param display: (str) 'DISPLAY' env variable, that is needed to be provided
    to perform action on linux platform.
    Defaults to 'localhost:99' since we have xvfb launched as service.
    :return: dict with key 'Success', results of redeploy, key 'Warnings',
    indicating if there were warnings during the process
    """
    logging.info(f"Redeploy process started")
    execution_log_lines = []
    execution_status = False
    warnings = False
    env = os.environ.copy()
    env['DISPLAY'] = display
    command = ['-profile', sas_conn_file_path, '-deploytype', deploytype,
               '-objects', *objects_list, '-sourcedir',
               deploy_dir, '-metarepository', metarepository, '-appservername',
               appservername, '-serverusername', serverusername,
               '-serverpassword', serverpassword, '-log', execution_log_path]
    command.insert(0, deployjob_jar_path)
    logging.debug(f"Command to be executed: {command}")
    try:
        redeploy_process = subprocess.run(command, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if redeploy_process.returncode == 0:
            logging.info(f"Redeploy process finished successfully.")
            execution_status = True
        elif redeploy_process.returncode == 4:
            execution_status = True
            warnings = True
            logging.warning(f"Redeploy process finished with warnings.")
            # execution_log_lines += redeploy_process.stdout
        else:
            # execution_log_lines += redeploy_process.stderr
            logging.error(f"Redeploy process finished with Errors! Check log!")
    except Exception as e:
        logging.error(f"Errors while performing redeploy {str(e)}")
        execution_log_lines += str(e)
    finally:
        logging.info(f"Log will be written to {execution_log_path}")
        with open(execution_log_path, 'a+') as log:
            log.writelines(execution_log_lines)
        return {'Success': execution_status, 'Warnings': warnings}


def export_objects(objects_list, log_path, sas_conn_file_path,
                   export_package_jar_path,
                   package_path):
    """
    Launches meta objects export process
    :param objects_list: list of metadata path strings
    :param log_path: path to lig file (it will be created by export process)
    :param sas_conn_file_path: path to .swa file with connection description
    :param export_package_jar_path: path to ExportPackage binary
    :param package_path: path to spk file, that will be created.
    :return: dict with key 'Success', results of export process, key 'Warnings',
    indicating if there were warnings during the process
    """
    execution_log_lines = []
    env = os.environ.copy()
    execution_status = False
    warnings = False
    logging.info(f"Starting metadata export process.")
    command = ['-profile', sas_conn_file_path, '-package', package_path,
               '-objects', *objects_list, '-log', log_path]
    if platform == "linux" or platform == "linux2":
        command.append('-disableX11')
    command.insert(0, export_package_jar_path)
    logging.debug(f"Command to be executed: {command}")
    try:
        export_process = subprocess.run(command, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # print(80 * '*')
        if export_process.returncode == 0:
            logging.info(f"Objects succesfully exported to {package_path}.")
            execution_status = True
        elif export_process.returncode == 4:
            # print(f'Export process finished with warnings.')
            execution_status = True
            warnings = True
            logging.warning(f"Warnings during export process. Check log: {log_path}")
            # logging.info(f"Objects exported to {package_path}.")
            # execution_log_lines += redeploy_process.stdout
        else:
            logging.error(f"Errors during export process. Check log: {log_path}")
            # execution_log_lines += redeploy_process.stderr
            # print(f'Export process finished with ERRORS.')
    except Exception as e:
        execution_log_lines += str(e)
        logging.error(f"Export process finished with ERRORS: {str(e)}")
    finally:
        logging.info(f"Log will be written to: {log_path}")
        with open(log_path, 'a') as log:
            log.writelines(execution_log_lines)
        return {'Success': execution_status, 'Warnings': warnings}


def backup_metadata(backup_script_path, comment, sas_conn_file_path, log_path):
    """
    Launches metadata backup
    :param backup_script_path: path to sas-backup-metadata script
    :param comment: comment(str) to be written on backup
    :param sas_conn_file_path: path to .swa file with connection description
    :param log_path: path to lig file
    :return: dict with key 'Success', results of backup process,
    key 'Backup_name', containing the backup name
    """
    logging.info(f"Full metadata backup started")
    execution_log_lines = []
    env = os.environ.copy()
    backup_success = False
    backup_name = None
    warnings = False
    command = [backup_script_path, '-comment', comment,
               '-profile', sas_conn_file_path, '-log', log_path]
    logging.debug(f"Command to run: {command}")
    try:
        backup_process = subprocess.run(command, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if backup_process.returncode == 0:
            backup_success = True
            backup_name = str(backup_process.stdout.readline()).split()[2]
            logging.info(f"Backup {backup_name} completed successfully.")
        elif backup_process.returncode == 4:
            backup_success = False
            backup_name = str(backup_process.stdout.readline()).split()[2]
            logging.warning(f"Backup {backup_name} completed with WARNINGS.")
            warnings = True
        else:
            # execution_log_lines += backup_process.stderr
            logging.warning(f"Backup NOT completed!")
    except Exception as e:
        execution_log_lines += str(e)
        logging.error(f"Errors during backup process! {str(e)}")
    finally:
        logging.info(f"Log will be written to {log_path}")
        with open(log_path, 'a+') as log:
            log.writelines(execution_log_lines)
        return {'Success': backup_success, 'Backup_name': backup_name}


def recover_metadata(recover_script_path, backup_name, comment,
                     sas_conn_file_path, log_path):
    """
    Launches metadata restore
    :param recover_script_path: path to sas-recover-metadata script
    :param backup_name: name of backup (actually, name of folder)
    :param comment: comment(str) to be written on restore
    :param sas_conn_file_path: path to .swa file with connection description
    :param log_path: path to lig file
    :return: dict with key 'Success', results of restore process,
    key 'Backup_name', containing the backup name
    """
    logging.info(f"Metadata recovery from backup {backup_name} started.")
    execution_log_lines = []
    env = os.environ.copy()
    restore_success = False
    warnings = False
    command = [recover_script_path, backup_name, '-comment', comment,
               '-profile', sas_conn_file_path, '-log', log_path]
    logging.info(f"Command to run: {command}")
    try:
        backup_process = subprocess.run(command, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        print(80 * '*')
        if backup_process.returncode == 0:
            restore_success = True
            logging.info(f"Restore from backup {backup_name} finished successfully")
        elif backup_process.returncode == 4:
            restore_success = False
            warnings = True
            logging.warning(f"Restore from backup {backup_name} finished with warnings")
            # execution_log_lines += redeploy_process.stdout
        else:
            logging.warning(f"Restore from backup {backup_name} not finished!")
            # execution_log_lines += redeploy_process.stderr
    except Exception as e:
        execution_log_lines += str(e)
        logging.error(f"Restore from backup {backup_name} not finished. ERRORS: {str(e)}")
    finally:
        logging.info(f"Log file will be written to {log_path}")
        with open(log_path, 'a') as log:
            log.writelines(execution_log_lines)
        return {'Success': restore_success}



from .sas.core import AbstractSAS
from .globals import sasma_sasobjsp_host, sasma_sasobjsp_port, sasma_workspace


class SASMA(AbstractSAS):
    """
    This class is a wrapper for saspy.SASSession.
    It hides all the 'tricky' parts of establishing connection with SASMA.
    """
    def __init__(self, user,
                       password,
                       sasobjsp_host=sasma_sasobjsp_host,
                       sasobjsp_port=sasma_sasobjsp_port,
                       appserver=sasma_workspace,
                       **kwargs):
        """
        Establishes sas connection.
        :param user: user, that is used to connect to SASMA.
        :param password: The password.
        :param sasobjsp_host: Host, on which SAS Object spawner is running.
        :param sasobjsp_port: Port, on which SAS Object spawner is running.
        :param kwargs:
        """
        self.platform = 'SASMA'
        self.env_type = 'Prod'
        super(SASMA, self).__init__(sasobjsp_host=sasobjsp_host,
                                    sasobjsp_port=sasobjsp_port,
                                    sasobjsp_user=user,
                                    sasobjsp_pass=password,
                                    appserver=self._is_appserver_valid(appserver),
                                    **kwargs)

    def _is_appserver_valid(self, appsever):
        allowed_appservers = [sasma_workspace]
        for existing_appserver in allowed_appservers:
            if existing_appserver.lower().startswith(appsever.lower()):
                return existing_appserver
        else:
            raise ValueError(f'Appserver must be one of: {allowed_appservers}')






from .sascodes import extract_lib_template
from .core import AbstractSAS


def extract_libraries(sasobjsp_user: str,
                      sasobjsp_pass: str,
                      sasobjsp_host: str,
                      sasobjsp_port: str,
                      appserver: str,
                      metadata_server: str,
                      metadata_port: str,
                      metadata_user: str,
                      metadata_password: str,
                      metadata_repo,
                      sas_session: AbstractSAS = None) -> list:

    get_all_libs_query = extract_lib_template.substitute(metadata_server=metadata_server,
                                                         metadata_port=metadata_port,
                                                         metadata_user=metadata_user,
                                                         metadata_password=metadata_password,
                                                         metadata_repo=metadata_repo)
    if not sas_session:
        sas_session = AbstractSAS(sasobjsp_port=sasobjsp_port,
                                  sasobjsp_host=sasobjsp_host,
                                  sasobjsp_user=sasobjsp_user,
                                  sasobjsp_pass=sasobjsp_pass,
                                  appserver=appserver)
    sas_session.submit(get_all_libs_query)
    df = sas_session.read_dataset(libname='WORK', tablename='Libraries_final')
    return df.to_json(orient='records')



""""
BACKWARD COMPITABILITY!!!!!
"""

#############################
# Sas support functions import
#############################
from .vault.vault import get_connection_from_vault, get_secret_from_vault
from .sas.core import get_objects_from_spk, check_import, redeploy_jobs, \
    launch_sas_code, import_spk, check_sas_code
#############################
# Platform connectors import
#############################
from .sasma import SASMA
from .sascs import SASCS, SASScoringVS221, SASScoringVS235
from .sasoa import SASOA
#############################
# BO Support functions import
#############################
from .bo import parse_dfx

#############################
# Win task parser
#############################
from .windows import XMLScheduledTask


""""
BACKWARD COMPITABILITY END.
"""


from string import Template


class SasCodeTemplate(Template):
    delimiter = '#'


extract_lib_template = SasCodeTemplate("""
%let metasrv="#{metadata_server}";
%let metaprt=#{metadata_port};
%let metausr="#{metadata_user}";
%let metapwd="#{metadata_password}";
%let metarep="#{metadata_repo}";

%macro libsfromrepo(mdrep=);
/*************************************************************/

/*%macro getMetaInfos(EXCELFILE,OUTPUTFORMAT);*/
  data metadata_libraries;
  length uri serveruri conn_uri domainuri conns_uri serv_uri propuri libname ServerContext AuthDomain path_schema
         usingpkguri type tableuri coluri $256 id $17
         desc $200 libref engine $8 isDBMS $1 DomainLogin  $32
		 Server_connection $100 Server_meta $50 TnsODBC $50 TnsODBC2 $50
			repo $50;
  /*keep libname desc libref engine ServerContext path_schema AuthDomain table colname
      coltype collen IsPreassigned IsDBMSLibname id Server_connection Server_meta TnsODBC TnsODBC2;*/
  keep libname desc libref engine ServerContext path_schema AuthDomain  
        IsPreassigned /*IsDBMSLibname*/ /*id*/ Server_connection Server_meta TnsODBC TnsODBC2 repo;

  nobj=.;
  n=1;
  uri='';
  serveruri='';
  conn_uri='';
  domainuri='';

         /***Determine how many libraries there are***/
  nobj=metadata_getnobj("omsobj:SASLibrary?@Id contains '.'",n,uri);
         /***Retrieve the attributes for all libraries, if there are any***/
  if n>0 then do n=1 to nobj;
    libname='';
    ServerContext='';
    AuthDomain='';
    desc='';
    libref='';
    engine='';
    isDBMS='';
    IsPreassigned='';
    IsDBMSLibname='';
    path_schema='';
    usingpkguri='';
    type='';
    id='';
	Server_connection='';
	Server_meta='';
	TnsODBC='';
	TnsODBC2='';

    nobj=metadata_getnobj("omsobj:SASLibrary?@Id contains '.'",n,uri);
    rc= metadata_getattr(uri, "Name", libname);
    rc= metadata_getattr(uri, "Desc", desc);
    rc= metadata_getattr(uri, "Libref", libref);
    rc= metadata_getattr(uri, "Engine", engine);
    rc= metadata_getattr(uri, "IsDBMSLibname", isDBMS);
    rc= metadata_getattr(uri, "IsDBMSLibname", IsDBMSLibname); 
    rc= metadata_getattr(uri, "IsPreassigned", IsPreassigned); 
    rc= metadata_getattr(uri, "Id", Id);

    /*** Get associated ServerContext ***/
    i=1;
    rc= metadata_getnasn(uri, "DeployedComponents", i, serveruri);
    if rc > 0 then rc2= metadata_getattr(serveruri, "Name", ServerContext);
    else ServerContext='';

    /*** If the library is a DBMS library, get the Authentication Domain
         associated with the DBMS connection credentials ***/
    if isDBMS="1" then do;
      i=1; 
      rc= metadata_getnasn(uri, "LibraryConnection", i, conn_uri);
      if rc > 0 then do;
        rc2= metadata_getnasn(conn_uri, "Domain", i, domainuri);
        if rc2 > 0 then do;
			rc3= metadata_getattr(domainuri, "Name", AuthDomain);
			rc4=metadata_getnasn(domainuri, "Connections", i, conns_uri);
			if rc4>0 then do;
				rc4=metadata_getattr(conns_uri, "Name", Server_connection);
				rc5=metadata_getnasn(conns_uri, "Source", i, serv_uri);
				if rc5>0 then do;
					rc5=metadata_getattr(serv_uri, "Name", Server_meta);
				end;
			end;
		end;
		rc2=metadata_getnasn(conn_uri, "Properties", i, propuri);
		if rc2 > 0 then do;
			rc3=metadata_getattr(propuri, "DefaultValue", TnsODBC);
		end;

		rc2_2=metadata_getnasn(conn_uri, "Properties", 2, propuri);
		if rc2_2 > 0 then do;
			rc3=metadata_getattr(propuri, "DefaultValue", TnsODBC2);
		end;

      end;
    end;

    /*** Get the path/database schema for this library ***/
    rc=metadata_getnasn(uri, "UsingPackages", 1, usingpkguri);
    if rc>0 then do;
      rc=metadata_resolve(usingpkguri,type,id);  
      if type='Directory' then 
        rc=metadata_getattr(usingpkguri, "DirectoryName", path_schema);
      else if type='DatabaseSchema' then 
        rc=metadata_getattr(usingpkguri, "SchemaName", path_schema);
      else path_schema="unknown";
    end;
  repo="&mdrep";
  output;
    
  end;
 
 run;

 proc append base=Libraries_final
 	data=metadata_libraries;
%mend;


options /*metaserver="sasmeta.dwhgridprod.imb.ru"*/
		metaserver=&metasrv
        metaport=&metaprt
        metauser=&metausr
        metapass=&metapwd
        metarepository=&metarep;

/*получаем список репозиториев SAS*/


filename myoutput temp;

proc metadata 
out=myoutput
header=full
   in="<GetRepositories>
   <Repositories/> 
   <!-- OMI_ALL (1) flag -->
   <Flags>1</Flags>
    <Options/> 
   </GetRepositories>";
run;


filename getrepos temp;
libname getrepos xmlv2 xmlfileref=myoutput automap=replace xmlmap=getrepos;
proc copy in=getrepos out=work;
run;

proc sql;
	delete from Repository
	where Repository_name in ('REPOSMGR','BILineage');
quit;


data run;
	set Repository;
	str = catt('options metaserver=',&metasrv,' metaport=',&metaprt,' metauser="',&metausr,'" metapass="',&metapwd,'" metarepository="', Repository_Name,'";');
	str_macro = catt('%libsfromrepo(mdrep=', Repository_Name,');');
	call execute(str);
	call execute(str_macro);
run;

proc sql;
	delete from Libraries_final
	where libname ='';
quit;
""")


import logging
import re
import subprocess

logger = logging.getLogger(__name__)


def get_odbc_sources() -> list:
    sources = []
    process = subprocess.run(['powershell', 'Get-OdbcDsn'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True, timeout=30)
    stdout = process.stdout
    stderr = process.stderr
    rc = process.returncode
    if rc == 0:
        logger.info(f'Successfully got odbc sources')
        logger.debug(f'stdout: "{stdout}"')
        pattern = r'^Name\s+:\s(?P<DSN_NAME>.*)$\n^DsnType\s+:\s(?P<DSN_TYPE>.*)$\n^Platform\s+:\s(?P<DSN_PLATFORM>.*)$\n^DriverName\s+:\s(?P<DSN_DRIVER_NAME>.*)$\n^Attribute\s+:\s(?P<DSN_ATTRIBUTE>.*)$'
        matches = re.finditer(pattern, string=stdout, flags=re.MULTILINE | re.VERBOSE)
        for match in matches:
            source = match.groupdict()
            logger.debug(f'Founded match: {source}')
            sources.append(source)
        logger.debug(f'Parse results: {sources}')
        return sources
    else:
        logger.error(f'Getting list of ODBC failed. Stderr: "{stderr}"')
        return sources


def kill_pid(pid: str):
    process = subprocess.run(['taskkill', '/PID', pid, '/T', '/F'], stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE, universal_newlines=True, timeout=30)
    stdout = process.stdout
    stderr = process.stderr
    rc = process.returncode
    return {"rc": rc, "stdout": stdout, "stderr": stderr}


def disable_scheduled_task(task_name: str = None):
    logger.info(f'Unscheduling of scheduled task {task_name} requested')
    process = subprocess.run(['powershell', 'Disable-ScheduledTask', '-TaskName', task_name], stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE, universal_newlines=True, timeout=30)
    stdout = process.stdout
    stderr = process.stderr
    rc = process.returncode
    if rc == 0:
        logger.info(f'Successfully unscheduled "{task_name}"')
        logger.debug(f'stdout: "{stdout}"')
        return {'rc': 0}
    else:
        logger.error(f'Unscheduling of task {task_name} failed. Stderr: "{stderr}"')
        return {'rc': rc}


def enable_scheduled_task(task_name: str = None):
    logger.info(f'Enabling of scheduled task {task_name} requested')
    process = subprocess.run(['powershell', 'Enable-ScheduledTask', '-TaskName', task_name], stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE, universal_newlines=True, timeout=30)
    stdout = process.stdout
    stderr = process.stderr
    rc = process.returncode
    if rc == 0:
        logger.info(f'Successfully scheduled "{task_name}"')
        logger.debug(f'stdout: "{stdout}"')
        return {'rc': 0}
    else:
        logger.error(f'Enabling of task {task_name} failed. Stderr: "{stderr}"')
        return {'rc': rc}


def delete_scheduled_task(task_name: str = None):
    logger.info(f'Deletion of scheduled task {task_name} requested')
    process = subprocess.run(['powershell', 'Unregister-ScheduledTask', '-TaskName', task_name, '-Confirm:$false'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True, timeout=30)
    stdout = process.stdout
    stderr = process.stderr
    rc = process.returncode
    if rc == 0:
        logger.info(f'Successfully deleted scheduled task "{task_name}"')
        logger.debug(f'stdout: "{stdout}"')
        return {'rc': 0}
    else:
        logger.error(f'Deletion of scheduled task {task_name} failed. Stderr: "{stderr}"')
        return {'rc': rc}


def parse_handle_out(handle_output: str) -> list:
    logger.debug(f'Started handle output parser')
    blockers = []
    pattern = r'^(?P<executable>\S+)\s+pid:\s(?P<pid>\d+)\s+type:\s(?P<type>\w+)\s+(?P<user>\S+)\s+(?P<handle>\w+):\s(?P<path>.*)$'
    matches = re.finditer(pattern, string=handle_output, flags=re.MULTILINE | re.VERBOSE)
    for match in matches:
        blocker = match.groupdict()
        logger.debug(f'Founded match: {blocker}')
        blockers.append(blocker)
    logger.debug(f'Parse results: {blockers}')
    return blockers


def unlock(path: str, handle_exe_path: str) -> dict:
    logger.info(f'Unlock of {path} requested')
    process = subprocess.run([handle_exe_path, '-u', path], stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE, universal_newlines=True, timeout=60)
    stdout = process.stdout
    stderr = process.stderr
    rc = process.returncode
    if rc == 0:
        logger.info(f'Successfully got list of processes, that uses "{path}"')
        logger.debug(f'stdout: "{stdout}"')
        logger.debug(f'Starting output parsing')
        blockers = parse_handle_out(handle_output=stdout)
        for blocker in blockers:
            logger.debug(f'Unblocking {blocker}')
            process = subprocess.run([handle_exe_path, '-c', blocker['handle'], '-p', blocker['pid'], '-y'],
                                     stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True,
                                     timeout=60)
            stdout = process.stdout
            stderr = process.stderr
            rc = process.returncode
            if rc == 0:
                logger.info(f'Process {blocker["pid"]} successfully killed.')
                logger.debug(f'stdout: "{stdout}"')
            else:
                logger.error(f'Unable to close handle for pid {blocker["pid"]}')
                logger.error(f'Following error occurred: "{stderr}"')
                break
        else:
            return {"rc": rc, "stdout": stdout, "stderr": stderr}
    else:
        logger.error(f'Unable to ger list of processes, that uses f"{path}"')
        logger.error(f'Following error occurred: "{stderr}"')
    return {"rc": rc, "stdout": stdout, "stderr": stderr}


def execute_xmla(xmla_query: str,  catalog: str, adomdclientdll: str, server: str) -> dict:
    if adomdclientdll:
        import clr
        clr.AddReference(adomdclientdll)
        clr.AddReference("System.Data")
        from Microsoft.AnalysisServices.AdomdClient import AdomdConnection
    else:
        logger.critical(f'AdomdClient.dll is required...')
        return {'rc': 1}
    logger.info(f'Starting to execute xmla query on {server}')
    logger.debug(f'Opening connection to {server}, catalog {catalog}')
    try:
        logger.debug(f'Initiating conn object')
        conn = AdomdConnection(f"Data Source={server};Catalog={catalog};")
        logger.debug(f'Conn object created, opening connection...')
        conn.Open()
        logger.debug(f'Conn opened...')
        if conn.State == 1:
            logger.info(f'Connection to {server} opened.')
        else:
            logger.error(f'Connection to {server} was not opened.')
            return {'rc': 1}
        logger.debug(f'Preparing connection to accept xmla query...')
        cmd = conn.CreateCommand()
        cmd.CommandText = xmla_query
        logger.debug(f'Xmla query is ready to be executed.')
        logger.info(f'Executing xmla query...')
        cmd.Execute()
        logger.info(f'Xmla query executed.')
        conn.Close()
        logger.info(f'Connection to {server} closed.')
        return {'rc': 0}
    except Exception as error:  # I know, i know...
        logger.error(f'Query executed with error: {str(error)}')
        return {'rc': 1, 'exception': str(error)}


import os

# Default Java
java_linux = '/usr/bin/java'  # fetch it from env vars???
java_windows = 'java'

# SAS globals...
all_obj_types = ['DeployedFlow', 'DeployedJob', 'ExternalFile', 'Folder',
                     'Library', 'Role', 'Server', 'StoredProcess', 'Table',
                     'User', 'Job']


#**********************************
# Default platfrom settings
###################################
# SASMA
###################################
sasma_sasobjsp_host = 'vs246.imb.ru'
sasma_sasobjsp_port = '8591'
sasma_workspace = 'SASApp_DI - Workspace Server'
##################################
# SAS Scoring
###################################
sascs_sasobjsp_host = 'vs2458.imb.ru'
sascs_sasobjsp_port = '8591'
sascs_workspace = 'DRKR - Workspace Server'
sascs_sasapp_workspace = 'SASApp - Workspace Server'
sascs_alloved_appservers = [sascs_sasapp_workspace]
#**********************************
sasvs235_sasobjsp_host = 'vs235.imb.ru'
sasvs235_sasobjsp_port = '8591'
sasvs235_drkr_workspace = 'DRKR - Workspace Server'
sasvs235_sasapp_workspace = 'SASApp - Workspace Server'
sasvs235_drkr_pm_workspace = 'DRKR_PM - Workspace Server'
sasvs235_drkr_fdp_workspace = 'DRKR_FDP - Workspace Server'
sasvs235_drkr_scu_workspace = 'DRKR_SCU - Workspace Server'
sasvs235_drkr_rt_workspace = 'DRKR_RT - Workspace Server'
sasvs235_drkr_rwo_workspace = 'DRKR_RWO - Workspace Server'
sasvs235_alloved_appservers = [sasvs235_drkr_workspace, sasvs235_sasapp_workspace, sasvs235_drkr_pm_workspace,
                               sasvs235_drkr_fdp_workspace, sasvs235_drkr_scu_workspace, sasvs235_drkr_rt_workspace,
                               sasvs235_drkr_rwo_workspace]
#**********************************
sasvs221_sasobjsp_host = 'vs221.imb.ru'
sasvs221_sasobjsp_port = '8591'
sasvs221_drkr_workspace = 'DRKR - Workspace Server'
sasvs221_sasapp_workspace = 'SASApp - Workspace Server'
sasvs221_dsur_workspace = 'DSUR - Workspace Server'
sasvs221_dfpc_workspace = 'DFPC - Workspace Server'
sasvs221_aso_workspace = 'ASO - Workspace Server'
sasvs221_alloved_appservers = [sasvs221_drkr_workspace, sasvs221_dsur_workspace, sasvs221_dfpc_workspace,
                               sasvs221_aso_workspace]
##################################
# SAS OA
###################################
sasoa_sasobjsp_host = 'vs990.imb.ru'
sasoa_sasobjsp_port = '8591'
sasoa_sasapp_workspace = 'SASApp - Workspace Server'
###################################
# HashicorpVault settings variables
###################################
hvac_url = 'https://hvault.intranet'
hvac_token_env_var = 'HVAC_TOKEN'
hvac_token = os.environ.get(hvac_token_env_var, default=None)
hvac_role_id_env_var = 'HVAC_ROLE_ID'
hvac_role_id = os.environ.get(hvac_role_id_env_var, default=None)
hvac_secret_env_var = 'HVAC_SECRET_ID'
hvac_secret_id = os.environ.get(hvac_secret_env_var, default=None)
hvac_connections_path = 'Data_Science/dlake/airflow/connections/'
hvac_connections_mountpoint = 'CDO/'



import hvac
import certifi
import logging
import os
from urllib.parse import parse_qsl, unquote, urlparse
from ..globals import hvac_url, hvac_token, hvac_secret_id, hvac_role_id, hvac_connections_path, \
    hvac_connections_mountpoint, hvac_token_env_var, hvac_secret_env_var, hvac_role_id_env_var

logging.getLogger(__name__)


def get_hvac_client(hvac_url: str = hvac_url,
                    namespace: str = None,
                    token: str = hvac_token,
                    hvac_role_id: str = hvac_role_id,
                    hvac_secret_id: str = hvac_secret_id,
                    ldap_login: str = None,
                    ldap_password: str = None,
                    verify=True,
                    **kwargs):
    if not token:
        token = os.environ.get(hvac_token_env_var, default=None)
    if not hvac_role_id:
        hvac_role_id = os.environ.get(hvac_role_id_env_var, default=None)
    if not hvac_secret_id:
        hvac_secret_id = os.environ.get(hvac_secret_env_var, default=None)
    try:
        if token:
            client = hvac.Client(url=hvac_url, token=token, namespace=namespace, verify=verify)
        elif hvac_role_id and hvac_secret_id:
            client = hvac.Client(url=hvac_url, namespace=namespace, verify=verify)
            client.auth.approle.login(hvac_role_id, hvac_secret_id, use_token=True)
        elif ldap_login and ldap_password:
            client = hvac.Client(url=hvac_url, namespace=namespace, verify=verify)
            client.auth.ldap.login(username=ldap_login, password=ldap_password)
        else:
            print(f"Something is missing, cannot authentificate.")
            raise ValueError
    except ConnectionError as e:
        print(f"Connection error. Probably, you don't have correct certificates/chain for {hvac_url}.\n"
              f"Certificates are located at {certifi.where()}\n"
              f"You can also use REQUESTS_CA_BUNDLE env variable to point to correct certs.\n"
              f"Full error: {str(e)}\n"
              f"Also, if you are some kind of naughty boy/girl/whatever you can pass verify=False as one of the args\n"
              f"Which is, obviously, very bad. Added it just for testing.\n")
        return
    return client


def _normalize_conn_type(conn_type):
    if conn_type == 'postgresql':
        conn_type = 'postgres'
    elif '-' in conn_type:
        conn_type = conn_type.replace('-', '_')
    return conn_type


def _parse_netloc_to_hostname(uri_parts):
    """Parse a URI string to get correct Hostname."""
    hostname = unquote(uri_parts.hostname or '')
    if '/' in hostname:
        hostname = uri_parts.netloc
        if "@" in hostname:
            hostname = hostname.rsplit("@", 1)[1]
        if ":" in hostname:
            hostname = hostname.split(":", 1)[0]
        hostname = unquote(hostname)
    return hostname


def _parse_from_uri(uri: str):
    uri_parts = urlparse(uri)
    conn_type = uri_parts.scheme
    conn_type = _normalize_conn_type(conn_type)
    host = _parse_netloc_to_hostname(uri_parts)
    quoted_schema = uri_parts.path[1:]
    schema = unquote(quoted_schema) if quoted_schema else quoted_schema
    login = unquote(uri_parts.username) if uri_parts.username else uri_parts.username
    password = unquote(uri_parts.password) if uri_parts.password else uri_parts.password
    port = uri_parts.port
    extra = ''
    if uri_parts.query:
        query = dict(parse_qsl(uri_parts.query, keep_blank_values=True))
        if '__extra__' in query:
            extra = query['__extra__']
        else:
            extra = query
    return {'conn_type': conn_type,
            'host': host,
            'schema': schema,
            'login': login,
            'password': password,
            'port': port,
            'extra': extra}


def get_secret_from_vault(vault_path: str, vault_mount_point: str, client: hvac.Client = None, **kwargs):
    if not client:
        client = get_hvac_client(**kwargs)
    creds = client.secrets.kv.v2.read_secret_version(path=vault_path, mount_point=vault_mount_point)
    creds = creds['data']['data']
    key = list(creds.keys())[0]  # EXTREMELY DANGER!
    value = list(creds.values())[0]  # EXTREMELY DANGER!
    return key, value


def get_connection_from_vault(connection_id: str,
                              connections_path: str = hvac_connections_path,
                              mount_point=hvac_connections_mountpoint,
                              client: hvac.Client = None, **kwargs):
    if not client:
        client = get_hvac_client(**kwargs)
    conn_path = f"{connections_path}/{connection_id}"
    response = client.secrets.kv.v2.read_secret_version(path=conn_path, mount_point=mount_point)
    conn_uri = response['data']['data']['conn_uri']
    return _parse_from_uri(conn_uri)



#!/usr/bin/env python

# Copyright (c) 2015, Yahoo Inc.
# Copyrights licensed under the BSD
# See the accompanying LICENSE.txt file for terms.

"""Parser for Windows Scheduled Task files
"""

import struct
import xml.etree.ElementTree as ET
import os
import datetime
import sys
import json

class Error(Exception):
    """Base exception."""


class FormatError(Error):
    """Error while parsing the job format."""


class XMLScheduledTask(object):

    def __init__(self, task_data=None):
        try:
            task_xml = ET.fromstring(task_data)
        except ET.ParseError:
            raise FormatError("Job file is not XML format")

        task_xml = self._strip_namespace(task_xml)

        # Load RegistrationInfo data
        self.uri = self._get_element_data(task_xml, "RegistrationInfo/URI")
        self.security_descriptor = self._get_element_data(task_xml, "RegistrationInfo/SecurityDescriptor")
        self.source = self._get_element_data(task_xml, "RegistrationInfo/Source")
        self.date = self._get_element_data(task_xml, "RegistrationInfo/Date")
        self.author = self._get_element_data(task_xml, "RegistrationInfo/Author")
        self.version = self._get_element_data(task_xml, "RegistrationInfo/Version")
        self.description = self._get_element_data(task_xml, "RegistrationInfo/Description")
        self.documentation = self._get_element_data(task_xml, "RegistrationInfo/Documentation")

        # Load Principal data
        self.principal_id = task_xml.find("Principals/Principal").get("id")
        self.user_id = self._get_element_data(task_xml, "Principals/Principal/UserId")
        self.logon_type = self._get_element_data(task_xml, "Principals/Principal/LogonType")
        self.group_id = self._get_element_data(task_xml, "Principals/Principal/GroupId")
        self.display_name = self._get_element_data(task_xml, "Principals/Principal/DisplayName")
        self.run_level = self._get_element_data(task_xml, "Principals/Principal/RunLevel")
        self.process_token_sid_type = self._get_element_data(task_xml, "Principals/Principal/ProcessTokenSidType")
        self.required_privileges = self._get_list_element_data(task_xml,
                                                               "Principals/Principal/RequiredPrivileges/Privilege")

        # Load Settings data
        self.allow_start_on_demand = self._get_element_data(task_xml, "AllowStartOnDemand")
        self.disallow_start_on_batteries = self._get_element_data(task_xml, "DisallowStartIfOnBatteries")
        self.stop_on_batteries = self._get_element_data(task_xml, "StopIfGoingOnBatteries")
        self.allow_hard_terminate = self._get_element_data(task_xml, "AllowHardTerminate")
        self.start_when_available = self._get_element_data(task_xml, "StartWhenAvailable")
        self.network_profile_name = self._get_element_data(task_xml, "NetworkProfileName")
        self.run_only_on_network = self._get_element_data(task_xml, "RunOnlyIfNetworkAvailable")
        self.wake_to_run = self._get_element_data(task_xml, "WakeToRun")
        self.enabled = self._get_element_data(task_xml, "Enabled")
        self.hidden = self._get_element_data(task_xml, "Hidden")
        self.delete_expired = self._get_element_data(task_xml, "DeleteExpiredTaskAfter")
        self.execution_time_limit = self._get_element_data(task_xml, "ExecutionTimeLimit")
        self.run_only_idle = self._get_element_data(task_xml, "RunOnlyIfIdle")
        self.unified_scheduling_engine = self._get_element_data(task_xml, "UseUnifiedSchedulingEngine")
        self.disallow_start_on_remote_app_session = self._get_element_data(task_xml, "DisallowStartOnRemoteAppSession")
        self.multiple_instances_policy = self._get_element_data(task_xml, "MultipleInstancesPolicy")
        self.priority = self._get_element_data(task_xml, "Priority")
        self.idle_duration = self._get_element_data(task_xml, "IdleSettings/Duration")
        self.idle_wait_timeout = self._get_element_data(task_xml, "IdleSettings/WaitTimeout")
        self.idle_stop_on_idle_end = self._get_element_data(task_xml, "IdleSettings/StopOnIdleEnd")
        self.idle_restart_on_idle = self._get_element_data(task_xml, "IdleSettings/RestartOnIdle")
        self.network_name = self._get_element_data(task_xml, "NetworkSettings/Name")
        self.network_id = self._get_element_data(task_xml, "NetworkSettings/Id")
        self.restart_on_fail_interval = self._get_element_data(task_xml, "RestartOnFailure/Interval")
        self.restart_on_fail_count = self._get_element_data(task_xml, "RestartOnFailure/Count")

        # Load Data data
        self.data = self._get_raw_xml(task_xml, "Data")

        # Load Triggers data
        self.triggers = self._get_triggers(task_xml)

        # Load Actions data
        self.actions = self._get_actions(task_xml)

    @staticmethod
    def _strip_namespace(data):
        if data.tag.startswith('{'):
            ns_length = data.tag.find('}')
            namespace = data.tag[0:ns_length+1]
        for element in data.getiterator():
            if element.tag.startswith(namespace):
                element.tag = element.tag[len(namespace):]
        return data

    @staticmethod
    def _get_element_data(data, path):
        try:
            return data.find(path).text
        except AttributeError:
            return ""

    @staticmethod
    def _get_list_element_data(data, path):
        values = []
        list_data = data.findall(path)
        if list_data is not None:
            for element in list_data:
                try:
                    values.append(element.text)
                except AttributeError:
                    pass
        return values

    @staticmethod
    def _get_list_tag_data(data, path):
        values = []
        list_data = data.findall(path)
        if list_data is not None:
            for element in list_data:
                values.append(element.tag)
        return values

    def _get_dict_data(self, data, path, fields):
        values = []
        value_list = data.findall(path)
        for value_element in value_list:
            value = {}
            for field in fields:
                value[field] = self._get_element_data(value_element, field)
            values.append(value)
        return values

    @staticmethod
    def _get_raw_xml(data, path):
        xml = data.find(path)
        if xml is not None:
            return ET.tostring(xml, encoding="utf-8")
        else:
            return ""

    def _get_triggers(self, data):
        triggers = []
        trigger_list = data.findall("Triggers/*")
        for trigger_element in trigger_list:
            trigger = {"Type": trigger_element.tag,
                       "Enabled": self._get_element_data(trigger_element, "Enabled"),
                       "StartBoundary": self._get_element_data(trigger_element, "StartBoundary"),
                       "EndBoundary": self._get_element_data(trigger_element, "EndBoundary"),
                       "ExecutionTimeLimit": self._get_element_data(trigger_element, "ExecutionTimeLimit"),
                       "Repetition": self._get_dict_data(trigger_element, "Repetition", ["Interval", "Duration",
                                                                                         "StopAtDurationEnd"])
                       }
            if trigger["Type"] == "BootTrigger":
                trigger["Delay"] = self._get_element_data(trigger_element, "Delay")
            elif trigger["Type"] == "RegistrationTrigger":
                trigger["Delay"] = self._get_element_data(trigger_element, "Delay")
            elif trigger["Type"] == "IdleTrigger":
                pass
            elif trigger["Type"] == "TimeTrigger":
                trigger["RandomDelay"] = self._get_element_data(trigger_element, "RandomDelay")
            elif trigger["Type"] == "EventTrigger":
                trigger["Subscription"] = self._get_element_data(trigger_element, "Subscription")
                trigger["Delay"] = self._get_element_data(trigger_element, "Delay")
                trigger["PeriodOfOccurrence"] = self._get_element_data(trigger_element, "PeriodOfOccurrence")
                trigger["NumberOfOccurrences"] = self._get_element_data(trigger_element, "NumberOfOccurrences")
                trigger["MatchingElement"] = self._get_element_data(trigger_element, "MatchingElement")
                trigger["ValueQueries"] = self._get_list_element_data(trigger_element, "ValueQueries/Value")
            elif trigger["Type"] == "LogonTrigger":
                trigger["UserId"] = self._get_element_data(trigger_element, "UserId")
                trigger["Delay"] = self._get_element_data(trigger_element, "Delay")
            elif trigger["Type"] == "SessionStateChangeTrigger":
                trigger["UserId"] = self._get_element_data(trigger_element, "UserId")
                trigger["Delay"] = self._get_element_data(trigger_element, "Delay")
                trigger["StateChange"] = self._get_element_data(trigger_element, "StateChange")
            elif trigger["Type"] == "CalendarTrigger":
                trigger["RandomDelay"] = self._get_element_data(trigger_element, "RandomDelay")
                if trigger_element.find("ScheduleByDay") is not None:
                    trigger["ScheduleByDay"] = self._get_dict_data(trigger_element, "ScheduleByDay", ["DaysInterval"])
                if trigger_element.find("ScheduleByWeek") is not None:
                    trigger["ScheduleByWeek"] = self._get_schedule_by_week(trigger_element)
                if trigger_element.find("ScheduleByMonth") is not None:
                    trigger["ScheduleByMonth"] = self._get_schedule_by_month(trigger_element)
                if trigger_element.find("ScheduleByMonthDayOfWeek") is not None:
                    trigger["ScheduleByMonthDayOfWeek"] = self._get_schedule_by_month_day_of_week(trigger_element)
            triggers.append(trigger)
        return triggers

    def _get_schedule_by_week(self, data):
        schedule = {"WeeksInterval": self._get_element_data(data, "ScheduleByWeek/WeeksInterval"),
                    "DaysOfWeek": self._get_list_tag_data(data, "ScheduleByWeek/DaysOfWeek/*")
                    }
        return schedule

    def _get_schedule_by_month(self, data):
        schedule = {"DaysOfMonth": self._get_list_element_data(data, "ScheduleByMonth/DaysOfMonth/Day"),
                    "Months": self._get_list_tag_data(data, "ScheduleByMonth/Months/*")
                    }
        return schedule

    def _get_schedule_by_month_day_of_week(self, data):
        schedule = {"Weeks": self._get_list_element_data(data, "ScheduleByMonthDayOfWeek/Weeks/Week"),
                    "DaysOfWeek": self._get_list_tag_data(data, "ScheduleByMonthDayOfWeek/DaysOfWeek/*"),
                    "Months": self._get_list_tag_data(data, "ScheduleByMonthDayOfWeek/Months/*")
                    }
        return schedule

    def _get_actions(self, data):
        actions = []
        action_list = data.findall("Actions/*")
        for action_element in action_list:
            action = {"Type": action_element.tag}
            if action["Type"] == "Exec":
                action["Command"] = self._get_element_data(action_element, "Command")
                action["Arguments"] = self._get_element_data(action_element, "Arguments")
                action["WorkingDirectory"] = self._get_element_data(action_element, "WorkingDirectory")
            elif action["Type"] == "ComHandler":
                action["ClassId"] = self._get_element_data(action_element, "ClassId")
                action["Data"] = self._get_raw_xml(action_element, "Data")
            elif action["Type"] == "SendEmail":
                action["Server"] = self._get_element_data(action_element, "Server")
                action["Subject"] = self._get_element_data(action_element, "Subject")
                action["To"] = self._get_element_data(action_element, "To")
                action["Cc"] = self._get_element_data(action_element, "Cc")
                action["Bcc"] = self._get_element_data(action_element, "Bcc")
                action["ReplyTo"] = self._get_element_data(action_element, "ReplyTo")
                action["From"] = self._get_element_data(action_element, "From")
                action["Body"] = self._get_element_data(action_element, "Body")
                action["Attachments"] = self._get_element_data(action_element, "Attachments/File")
                action["Headers"] = self._get_dict_data(action_element, "Headers/HeaderField", ["Name", "Value"])
            elif action["Type"] == "ShowMessage":
                action["Title"] = self._get_element_data(action_element, "Title")
                action["Body"] = self._get_element_data(action_element, "Body")
            actions.append(action)
        return actions

    def parse(self):
        # Returns a dictionary containing all the fields of a XML task
        task = {
            "task_type": "XML",
            "uri": self.uri,
            "security_descriptor": self.security_descriptor,
            "source": self.source,
            "date": self.date,
            "author": self.author,
            "version": self.version,
            "description": self.description,
            "documentation": self.documentation,
            "principal_id": self.principal_id,
            "user_id": self.user_id,
            "logon_type": self.logon_type,
            "group_id": self.group_id,
            "display_name": self.display_name,
            "run_level": self.run_level,
            "process_token_sid_type": self.process_token_sid_type,
            "required_privileges": self.required_privileges,
            "allow_start_on_demand": self.allow_start_on_demand,
            "disallow_start_on_batteries": self.disallow_start_on_batteries,
            "stop_on_batteries": self.stop_on_batteries,
            "allow_hard_terminate": self.allow_hard_terminate,
            "start_when_available": self.start_when_available,
            "network_profile_name": self.network_profile_name,
            "run_only_on_network": self.run_only_on_network,
            "wake_to_run": self.wake_to_run,
            "enabled": self.enabled,
            "hidden": self.hidden,
            "delete_expired": self.delete_expired,
            "execution_time_limit": self.execution_time_limit,
            "run_only_idle": self.run_only_idle,
            "unified_scheduling_engine": self.unified_scheduling_engine,
            "disallow_start_on_remote_app_session": self.disallow_start_on_remote_app_session,
            "multiple_instances_policy": self.multiple_instances_policy,
            "priority": self.priority,
            "idle_duration": self.idle_duration,
            "idle_wait_timeout": self.idle_wait_timeout,
            "idle_stop_on_idle_end": self.idle_stop_on_idle_end,
            "idle_restart_on_idle": self.idle_restart_on_idle,
            "network_name": self.network_name,
            "network_id": self.network_id,
            "restart_on_fail_interval": self.restart_on_fail_interval,
            "restart_on_fail_count": self.restart_on_fail_count,
            "data": self.data,
            "triggers": self.triggers,
            "actions": self.actions
        }
        return task


class BinaryScheduledTask(object):
    PRODUCT_VERSIONS = {
        0x0400: "Windows NT 4.0",
        0x0500: "Windows 2000",
        0x0501: "Windows XP",
        0x0600: "Windows Vista",
        0x0601: "Windows 7",
        0x0602: "Windows 8",
        0x0603: "Windows 8.1"
    }

    FILE_VERSIONS = {
        0x0001: "Ver. 1"
    }

    PRIORITY = {
        0b100000: "NORMAL_PRIORITY_CLASS",
        0b1000000: "IDLE_PRIORITY_CLASS",
        0b10000000: "HIGH_PRIORITY_CLASS",
        0b100000000: "REALTIME_PRIORITY_CLASS"
    }

    STATUS = {
        0x00041300: "SCHED_S_TASK_READY",
        0x00041301: "SCHED_S_TASK_RUNNING",
        0x00041305: "SCHED_S_TASK_NOT_SCHEDULED"
    }

    TASK_FLAGS = {
        0b1: "TASK_FLAG_INTERACTIVE",
        0b10: "TASK_FLAG_DELETE_WHEN_DONE",
        0b100: "TASK_FLAG_DISABLED",
        0b10000: "TASK_FLAG_START_ONLY_IF_IDLE",
        0b100000: "TASK_FLAG_KILL_ON_IDLE_END",
        0b1000000: "TASK_FLAG_DONT_START_IF_ON_BATTERIES",
        0b10000000: "TASK_FLAG_KILL_IF_GOING_ON_BATTERIES",
        0b100000000: "TASK_FLAG_RUN_ONLY_IF_DOCKED",
        0b1000000000: "TASK_FLAG_HIDDEN",
        0b10000000000: "TASK_FLAG_RUN_IF_CONNECTED_TO_INTERNET",
        0b100000000000: "TASK_FLAG_RESTART_ON_IDLE_RESUME",
        0b1000000000000: "TASK_FLAG_SYSTEM_REQUIRED",
        0b10000000000000: "TASK_FLAG_RUN_ONLY_IF_LOGGED_ON",
        0b1000000000000000000000000: "TASK_APPLICATION_NAME"
    }

    EXITCODE = {
        0x0: "S_OK",
        0x1: "S_FALSE",
        0x80000002: "E_OUTOFMEMORY",
        0x80000009: "E_ACCESSDENIED",
        0x80000003: "E_INVALIDARG",
        0x80000008: "E_FAIL",
        0x8000FFFF: "E_UNEXPECTED",
        0x00041300: "SCHED_S_TASK_READY",
        0x00041301: "SCHED_S_TASK_RUNNING",
        0x00041302: "SCHED_S_TASK_DISABLED",
        0x00041303: "SCHED_S_TASK_HAS_NOT_RUN",
        0x00041304: "SCHED_S_TASK_NO_MORE_RUNS",
        0x00041305: "SCHED_S_TASK_NOT_SCHEDULED",
        0x00041306: "SCHED_S_TASK_TERMINATED",
        0x00041307: "SCHED_S_TASK_NO_VALID_TRIGGERS",
        0x00041308: "SCHED_S_EVENT_TRIGGER",
        0x80041309: "SCHED_E_TRIGGER_NOT_FOUND",
        0x8004130A: "SCHED_E_TASK_NOT_READY",
        0x8004130B: "SCHED_E_TASK_NOT_RUNNING",
        0x8004130C: "SCHED_E_SERVICE_NOT_INSTALLED",
        0x8004130D: "SCHED_E_CANNOT_OPEN_TASK",
        0x8004130E: "SCHED_E_INVALID_TASK",
        0x8004130F: "SCHED_E_ACCOUNT_INFORMATION_NOT_SET",
        0x80041310: "SCHED_E_ACCOUNT_NAME_NOT_FOUND",
        0x80041311: "SCHED_E_ACCOUNT_DBASE_CORRUPT",
        0x80041312: "SCHED_E_NO_SECURITY_SERVICES",
        0x80041313: "SCHED_E_UNKNOWN_OBJECT_VERSION",
        0x80041314: "SCHED_E_UNSUPPORTED_ACCOUNT_OPTION",
        0x80041315: "SCHED_E_SERVICE_NOT_RUNNING",
        0x80041316: "SCHED_E_UNEXPECTEDNODE",
        0x80041317: "SCHED_E_NAMESPACE",
        0x80041318: "SCHED_E_INVALIDVALUE",
        0x80041319: "SCHED_E_MISSINGNODE",
        0x8004131A: "SCHED_E_MALFORMEDXML",
        0x0004131B: "SCHED_S_SOME_TRIGGERS_FAILED",
        0x0004131C: "SCHED_S_BATCH_LOGON_PROBLEM",
        0x8004131D: "SCHED_E_TOO_MANY_NODES",
        0x8004131E: "SCHED_E_PAST_END_BOUNDARY",
        0x8004131F: "SCHED_E_ALREADY_RUNNING",
        0x80041320: "SCHED_E_USER_NOT_LOGGED_ON",
        0x80041321: "SCHED_E_INVALID_TASK_HASH",
        0x80041322: "SCHED_E_SERVICE_NOT_AVAILABLE",
        0x80041323: "SCHED_E_SERVICE_TOO_BUSY",
        0x80041324: "SCHED_E_TASK_ATTEMPTED",
        0x00041325: "SCHED_S_TASK_QUEUED",
        0x80041326: "SCHED_E_TASK_DISABLED",
        0x80041327: "SCHED_E_TASK_NOT_V1_COMPAT",
        0x80041328: "SCHED_E_START_ON_DEMAND"
    }

    TRIGGER_FLAGS = {
        0b1: "TASK_TRIGGER_FLAG_HAS_END_DATE",
        0b10: "TASK_TRIGGER_FLAG_KILL_AT_DURATION_END",
        0b100: "TASK_TRIGGER_FLAG_DISABLED"
    }

    def __init__(self, task_data=None):
        # Load fixed length sections
        self.product_version = struct.unpack("<H", task_data[0:2])[0]
        self.file_version = struct.unpack("<H", task_data[2:4])[0]

        if self.product_version not in BinaryScheduledTask.PRODUCT_VERSIONS:
            raise FormatError("Job file is not binary format")

        if self.file_version not in BinaryScheduledTask.FILE_VERSIONS:
            raise FormatError("Job file is not binary format")

        self.uuid = self._get_uuid(task_data)
        self.application_name_offset = struct.unpack("<H", task_data[20:22])[0]
        self.trigger_offset = struct.unpack("<H", task_data[22:24])[0]
        self.error_retry_count = struct.unpack("<H", task_data[24:26])[0]
        self.error_retry_interval = struct.unpack("<H", task_data[26:28])[0]
        self.idle_deadline = struct.unpack("<H", task_data[28:30])[0]
        self.idle_wait = struct.unpack("<H", task_data[30:32])[0]
        self.priority = struct.unpack("<I", task_data[32:36])[0]
        self.maximum_run_time = struct.unpack("<I", task_data[36:40])[0]
        self.exitcode = struct.unpack("<i", task_data[40:44])[0]
        self.status = struct.unpack("<I", task_data[44:48])[0]
        self.flags = struct.unpack("<I", task_data[48:52])[0]
        self.lastrun = self._get_last_run(task_data[52:68])

        # Load variable length section
        self.running_instance_count = struct.unpack("<H", task_data[68:70])[0]

        self.read_offset = self.application_name_offset
        self.application_name = self._get_offset_length_data(task_data, unicode_string=True)
        self.parameters = self._get_offset_length_data(task_data, unicode_string=True)
        self.working_dir = self._get_offset_length_data(task_data, unicode_string=True)
        self.author = self._get_offset_length_data(task_data, unicode_string=True)
        self.comment = self._get_offset_length_data(task_data, unicode_string=True)
        self.user_data = self._get_offset_length_data(task_data)
        self.reserved_data = self._get_offset_length_data(task_data)
        self.triggers = self._get_triggers(task_data)

        # Load optional signature
        try:
            self.signature = self._get_signature(task_data)
        except struct.error:
            self.signature = None

    @staticmethod
    def _get_uuid(data):
        return data[4:20]

    @staticmethod
    def _get_last_run(data):
        year = struct.unpack("<H", data[0:2])[0]
        month = struct.unpack("<H", data[2:4])[0]
        day = struct.unpack("<H", data[6:8])[0]
        hour = struct.unpack("<H", data[8:10])[0]
        minute = struct.unpack("<H", data[10:12])[0]
        second = struct.unpack("<H", data[12:14])[0]
        milliseconds = struct.unpack("<H", data[14:16])[0]
        return datetime.datetime(year, month, day, hour, minute, second, milliseconds * 1000)

    def _get_offset_length_data(self, data, unicode_string=False):
        # Read the 2 byte length of data and then return the data from offset, for unicode strings the length is the
        # number of characters so it
        value = ""
        value_begin = self.read_offset + 2
        value_length = int(struct.unpack("<H", data[self.read_offset:value_begin])[0])
        if value_length:
            if unicode_string:
                value_length *= 2
            value_end = value_begin + value_length
            value = data[value_begin:value_end]
        self.read_offset += value_length + 2
        if value and unicode_string:
            value = value[0:value_length-2]
        return value

    def _get_triggers(self, data):
        triggers = []
        triggers_begin = self.trigger_offset + 2
        triggers_length = struct.unpack("<H", data[self.trigger_offset:triggers_begin])[0]
        self.read_offset = triggers_begin
        for _ in range(0, triggers_length):
            trigger = {}
            trigger_data = data[self.read_offset:self.read_offset + 48]
            trigger["trigger_size"] = struct.unpack("<H", trigger_data[0:2])[0]
            trigger["reserved"] = struct.unpack("<H", trigger_data[2:4])[0]
            trigger["begin_year"] = struct.unpack("<H", trigger_data[4:6])[0]
            trigger["begin_month"] = struct.unpack("<H", trigger_data[6:8])[0]
            trigger["begin_day"] = struct.unpack("<H", trigger_data[8:10])[0]
            trigger["end_year"] = struct.unpack("<H", trigger_data[10:12])[0]
            trigger["end_month"] = struct.unpack("<H", trigger_data[12:14])[0]
            trigger["end_day"] = struct.unpack("<H", trigger_data[14:16])[0]
            trigger["start_hour"] = struct.unpack("<H", trigger_data[16:18])[0]
            trigger["start_minute"] = struct.unpack("<H", trigger_data[18:20])[0]
            trigger["minutes_duration"] = struct.unpack("<I", trigger_data[20:24])[0]
            trigger["minutes_interval"] = struct.unpack("<I", trigger_data[24:28])[0]
            trigger["flags"] = struct.unpack("<I", trigger_data[28:32])[0]
            trigger["trigger_type"] = struct.unpack("<I", trigger_data[32:36])[0]
            trigger["trigger_specific0"] = struct.unpack("<H", trigger_data[36:38])[0]
            trigger["trigger_specific1"] = struct.unpack("<H", trigger_data[38:40])[0]
            trigger["trigger_specific2"] = struct.unpack("<H", trigger_data[40:42])[0]
            trigger["padding"] = struct.unpack("<H", trigger_data[42:44])[0]
            trigger["reserved2"] = struct.unpack("<H", trigger_data[44:46])[0]
            trigger["reserved3"] = struct.unpack("<H", trigger_data[46:48])[0]
            triggers.append(trigger)
            self.read_offset += trigger["trigger_size"]
        return triggers

    def _get_signature(self, data):
        signature = {"version": struct.unpack("<H", data[self.read_offset:self.read_offset + 2])[0],
                     "min_client_version": struct.unpack("<H", data[self.read_offset + 2:self.read_offset + 4])[0],
                     "data": data[self.read_offset + 4:self.read_offset + 64]
                     }
        return signature

    def parse(self):
        # Returns a dictionary containing all the fields of a binary task
        flags = []
        for flag in self.TASK_FLAGS:
            if flag & self.flags:
                flags.append(self.TASK_FLAGS[flag])

        triggers = self.triggers
        for trigger in triggers:
            flags = []
            for flag in self.TRIGGER_FLAGS:
                if flag & trigger["flags"]:
                    flags.append(self.TRIGGER_FLAGS[flag])
            trigger["flags"] = flags

        data1 = struct.unpack("<I", self.uuid[0:4])[0]
        data2 = struct.unpack("<H", self.uuid[4:6])[0]
        data3 = struct.unpack("<H", self.uuid[6:8])[0]
        data4 = struct.unpack(">HHHH", self.uuid[8:16])

        uuid = "%08X-%04X-%04X-%04X-%04X%04X%04X" % (data1, data2, data3,
                data4[0], data4[1], data4[2], data4[3])

        task = {
            "task_type": "Binary",
            "application_name": self.application_name.decode('utf-16'),
            "author": self.author.decode('utf-16'),
            "comment": self.comment.decode('utf-16'),
            "error_retry_count": self.error_retry_count,
            "error_retry_interval": self.error_retry_interval,
            "exitcode": self.EXITCODE[self.exitcode],
            "file_version": self.FILE_VERSIONS[self.file_version],
            "flags": flags,
            "idle_deadline": self.idle_deadline,
            "idle_wait": self.idle_wait,
            "lastrun": str(self.lastrun),
            "maximum_run_time": self.maximum_run_time,
            "parameters": self.parameters.decode('utf-16'),
            "priority": self.PRIORITY[self.priority],
            "product_version": self.PRODUCT_VERSIONS[self.product_version],
            "reserved_data": self.reserved_data,
            "running_instance_count": self.running_instance_count,
            "signature": self.signature,
            "status": self.STATUS[self.status],
            "triggers": triggers,
            "user_data": self.user_data,
            "uuid": uuid,
            "working_dir": self.working_dir.decode('utf-16')
        }
        return task


def read_task(file_data):
    # Returns a task object based on file data provided
    try:
        task_object = XMLScheduledTask(file_data)
    except FormatError:
        try:
            task_object = BinaryScheduledTask(file_data)
        except FormatError:
            return None
    return task_object

if __name__ == "__main__":
    # Enables command line usage.
    if len(sys.argv) <= 1:
        print ("Usage: %s <file>" % sys.argv[0])
        sys.exit(-1)

    filename = sys.argv[1]

    if os.path.isfile(filename):
        fd = open(filename, "rb")
        file_data = fd.read()
        fd.close()
        try:
            task = XMLScheduledTask(file_data)
        except FormatError:
            try:
                task = BinaryScheduledTask(file_data)
            except FormatError:
                print ("File is not XML or Binary job format")
                sys.exit(-1)
        print(json.dumps(task.parse(), indent=4))



import sqlparse
import csv
import os
import zipfile
import xml.etree.ElementTree as ET
from xml.etree import ElementTree
from sqlparse.sql import IdentifierList, Identifier
from sqlparse.tokens import Keyword, DML
import itertools
import re

ns = {'xmi': "http://www.omg.org/XMI",
      'xsi':  "http://www.w3.org/2001/XMLSchema-instance",
      'datafoundation': "http://com.businessobjects.mds.datafoundation"}


def _is_subselect(parsed):
    if not parsed.is_group:
        return False
    for item in parsed.tokens:
        if item.ttype is DML and item.value.upper() == 'SELECT':
            return True
    return False


def _extract_from_part(parsed):
    from_seen = False
    for item in parsed.tokens:
        if item.is_group:
            for x in _extract_from_part(item):
                yield x
        if from_seen:
            if _is_subselect(item):
                for x in _extract_from_part(item):
                    yield x
            elif item.ttype is Keyword and item.value.upper() in ['ORDER', 'GROUP', 'BY', 'HAVING', 'GROUP BY', 'ORDER BY']:
                from_seen = False
                StopIteration
            else:
                yield item
        if item.ttype is Keyword and item.value.upper() == 'FROM':
            from_seen = True


def _extract_table_identifiers(token_stream):
    for item in token_stream:
        if isinstance(item, IdentifierList):
            for identifier in item.get_identifiers():
                value = identifier.value.replace('"', '').lower()
        elif isinstance(item, Identifier):
            value = item.value.replace('"', '').lower()
            yield value


def _clean_table_name(table_name):
    import re
    table_name = table_name.strip()
    if re.match(r'^[\w\.]+\s+(as\s+)?\w+$', table_name):
        table_name = table_name.split()[0]
    return table_name


def _extract_tables(sql):
    extracted_tables = []
    statements = list(sqlparse.parse(sql))
    for statement in statements:
        if statement.get_type() != 'UNKNOWN':
            stream = _extract_from_part(statement)
            extracted_tables.append(set(list(_extract_table_identifiers(stream))))
        else:
            print('Wrong sql!')
            exit(1)
    ret1 = list(itertools.chain(*extracted_tables))
    ret2 = []
    for x in ret1:
        """
        sometimes sql is not parsed correctly (mainly subselects), so we need to 
        delete some obvious junk.
        """
        if '.' not in x:
            pass
        elif '(SELECT' in x.upper():
            pass
        elif ' AS ' in x.upper():
            ret2.append(x.split()[0])
        else:
            ret2.append(x.upper())
    ret3 = [_clean_table_name(tn) for tn in ret2]
    return list(set(ret3)) # make unique


def parse_dfx(dfx_path: str, results_dir: str):
    if not os.path.exists(dfx_path):
        return
    if not os.path.exists(results_dir):
        os.makedirs(results_dir)
    csv_res_file = os.path.join(results_dir, os.path.basename(dfx_path).replace('.dfx', '.csv'))
    dfx = zipfile.ZipFile(dfx_path)
    xml = ''
    for filename in dfx.namelist():
        if filename.endswith('.dfx'):
            xml = filename
            break
    if not xml:
        return 
    mapping = ET.parse(dfx.open(xml))
    root = mapping.getroot()
    parse_results = []
    namespaces = dict([
        node for _, node in ElementTree.iterparse(
            dfx.open(xml), events=['start-ns'])
    ])
    for tag in root.findall('tables', namespaces):
        tables = []
        xsi_type = tag.get('{http://www.w3.org/2001/XMLSchema-instance}type')
        type = tag.get('type')
        owner = tag.get('owner')
        businessname = tag.get('businessName')
        qualifier = tag.get('qualifier')
        expression_raw = tag.get('expression')
        keywords = ''
        if expression_raw and xsi_type == 'datafoundation:DerivedTable':
            expression_clean = sqlparse.format(expression_raw, reindent=True, keyword_case='upper', strip_comments=True)
            keywrds = re.findall(r'.*(?P<keyword>@\w+)\((?P<value>[^)]*)\)',
                                 string=expression_clean,
                                 flags=re.MULTILINE | re.VERBOSE)
            for keyword in keywrds:
                keywords += str(keyword)
            tables = _extract_tables(expression_clean)
            if not tables:
                if 'FROM DUAL' in expression_clean.upper():
                    tables = ['DUAL']
                else:
                    """
                    Process the condition, where no tables was extracted by regular extracter.
                    This is usually happening, when there is a join without join (from a,b where a.a=b.b)

                    """
                    findings = re.findall(r'FROM(?P<TABLES>([A-z0-9\.,\s]+))WHERE',
                                          string=expression_clean,
                                          flags=re.MULTILINE | re.VERBOSE)
                    for tup in findings:
                        for elem in tup:
                            if '\n' in elem:
                                elems = elem.replace(',', '')
                                elems = elem.split('\n')
                            else:
                                elems = elem.split(',')
                            for table in elems:
                                table = table.strip()
                                table = table.replace(',', '')
                                if table:
                                    table = table.split()[0].upper()
                                    tables.append(table)
            tables = list(set(tables))
            with open(os.path.join(results_dir, businessname + '.sql'), 'w') as queryfile:
                queryfile.write(expression_clean)
        if tables:
            for table in tables:
                object_info = {'XSI_TYPE': xsi_type, 'TYPE': type, 'OWNER': owner, 'BUSINESSNAME': businessname,
                               'TABLES': table, 'QUALIFIER': qualifier, 'KEYWORDS': keywords}
                parse_results.append(object_info)
        else:
            object_info = {'XSI_TYPE': xsi_type, 'TYPE': type, 'OWNER': owner, 'BUSINESSNAME': businessname,
                           'TABLES': None, 'QUALIFIER': qualifier, 'KEYWORDS': keywords}
            parse_results.append(object_info)

    fieldnames = ['XSI_TYPE', 'TYPE', 'OWNER', 'BUSINESSNAME', 'TABLES', 'QUALIFIER', 'KEYWORDS']
    with open(csv_res_file, 'w', newline='') as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames, dialect=csv.excel)
        writer.writeheader()
        for row in parse_results:
            writer.writerow(row)
    print(f'Extracted to {csv_res_file}')
    return results_dir
